require("defines")
-- debug_status = 1
debug_file = "mtools-deb.txt"
debug_mod_name = "MiningTools"
require("utils")
require("config")

-- status of water injector
local status_checking = -1  -- checking if zone already exploited
local status_preparing = 0  -- preparing injections
local status_injecting = 1  -- injecting and revealing ores
local status_finished = 2   -- process finished

--------------------------------------------------------------------------------------
function init_globals()
	global.ticks = global.ticks or 0
	global.dynamites = global.dynamites or {}
	global.injectors = global.injectors or {}
	global.zones_used = global.zones_used or {}
	global.ores = global.ores or {}
	global.fluids = global.fluids or {}
end

script.on_init( 
	function()
		init_globals ()
		update_ores_fluids()
	end
)

--------------------------------------------------------------------------------------
function on_configuration_changed(data)
	if data.mod_changes ~= nil then
		local mt_changes = data.mod_changes["MiningTools"]
		if mt_changes ~= nil then
			debug_print( "on_configuration_changed: MiningTools " .. tostring(mt_changes.old_version) .. " to " .. tostring(mt_changes.new_version) )
			init_globals()
		end
		
		-- update any other mod could result in a newx ores/fluids list...
		update_ores_fluids()
	end
end

script.on_configuration_changed(on_configuration_changed)

--------------------------------------------------------------------------------------
function on_creation( event )
	local ent = event.created_entity
	
	if ent.name == "water-injector" then
		debug_print( "injector created" )

		local inj = 			
		{
			entity = ent, 
			status=status_checking, 
			max_oils = math.floor(#global.fluids * injector_max_oils * (2+math.random())/3 ),
			cycles = injector_cycles,
			nb_try_ores = 0,
			n_boom = 0,
			centers={},
			zones={}, 
			elements={},
		}

		table.insert( global.injectors, inj	)
		
		if check_zone_used( inj, ent.position ) then
			inj.status = status_preparing
		else
			if event.player_index then -- test if player or robot
				game.get_player(event.player_index).print( {"already-injected"} )
			end
			inj.status = status_finished
			ent.active = false
			ent.operable = false
		end
		
		debug_print( "injectors=" .. #global.injectors )
		
	elseif string.sub(ent.name,1,9) == "dynamite-" then
		--local pos = event.created_entity.position
		local dyna = ent.name
		local ore = string.sub(ent.name,10)
		ent.operable = false
		
		table.insert( global.dynamites,
			{
				entity = ent,
				--position = pos,
				ore = ore,
				countdown = dynamite_countdown,
			}
		)
		
		debug_print( "dynamites=" .. #global.dynamites )
	end
end

script.on_event(defines.events.on_built_entity, on_creation )
script.on_event(defines.events.on_robot_built_entity, on_creation )

--------------------------------------------------------------------------------------
function on_destruction( event )
	local ent = event.entity
	
	if ent.name == "water-injector" then
		debug_print( "injector destroyed" )
		local i, inj
		
		for i, inj in pairs(global.injectors) do
			if inj.entity == ent then
				table.remove( global.injectors, i )
				break
			end
		end
		
		debug_print( "injectors=" .. #global.injectors )
	elseif string.sub(ent.name,1,9) == "dynamite-" then
		debug_print( "dynamite destroyed" )
		local i, dyna
		
		for i, dyna in pairs(global.dynamites) do
			if dyna.entity == ent then
				table.remove( global.dynamites, i )
				break
			end
		end
		
		debug_print( "dynamites=" .. #global.dynamites )
	end
end

script.on_event(defines.events.on_entity_died, on_destruction )
script.on_event(defines.events.on_robot_pre_mined, on_destruction )
script.on_event(defines.events.on_preplayer_mined_item, on_destruction )

--------------------------------------------------------------------------------------
function on_tick(event)
	if global.ticks == 0 then
		global.ticks = math.floor(injector_ticks * (10+math.random(-3,3))/10)
		local i=1
		local inj
		
		-- manage injectors
		
		while i <= #global.injectors do
			inj = global.injectors[i]
			local ent = inj.entity

			if ent and ent.valid then
				if inj.status == status_checking then
					-- check if zone already exploited
					-- if check_zone_used( inj, ent.position ) then
						-- inj.status = status_preparing
					-- else
						-- inj.status = status_finished
						-- ent.active = false
						-- ent.operable = false
						-- break
					-- end
				elseif inj.status == status_preparing then
					-- prepare injections
					local inv = ent.get_output_inventory()
					
					if inv then
						local nb_inj = inv.get_item_count("water-injection")
						
						if nb_inj >= injector_injections_threshold then
							debug_print( "inj=" .. nb_inj )
							inv.remove( {name="water-injection", count=injector_injections_threshold} )
							inj.status = status_injecting
							ent.active = false
							initiate_resources_zones( inj, ent.position )
							inj.nb_try_ores = math.floor( 2 + #inj.zones / inj.cycles )
							debug_print( "nb try ores=" .. inj.nb_try_ores)
							table.insert( global.zones_used, ent.position )
						end
					end
				elseif inj.status == status_injecting then
					-- injecting and revealing						
					reveal_resources_progressive( inj )
				end
			else
				table.remove(global.injectors, i)
				i=i-1
			end
			i=i+1
		end
	else
		global.ticks = global.ticks - 1
	end
	
	if #global.dynamites > 0 then
		-- manage dynamite explosions
		local  i=1
		local dyna
		while i <= #global.dynamites do
			dyna = global.dynamites[i]
			if dyna.countdown <= 0 then
				local ent = dyna.entity
				local surf = ent.surface
				local pos = ent.position
				local ore = dyna.ore
				local cancel_dynamite = false
				
				local ents = surf.find_entities_filtered({area = square_area(pos,0.4), type = "resource"})
				for _, ent in pairs(ents) do
					if is_fluid(ent.name) and ore ~= "wipe" then
						-- do not overwrite fluid resource with another dynamite, unless it is a "wipe" one
						cancel_dynamite = true
					else
						ent.destroy()
					end
				end
				
				if ore == "wipe" and #ents == 0 then
					-- do not wipe if no resource here
					cancel_dynamite = true
				end

				if not cancel_dynamite then
					ent = dyna.entity
					
					if ent and ent.valid then
						ent.destroy()
					end

					surf.create_entity({name= "explosion", position=pos})

					if ore ~= "wipe" then				
						if is_fluid(ore) then
							surf.create_entity({name=ore, amount=75 * math.random(dynamite_oil_min_yield,dynamite_oil_max_yield), position=pos})
						else
							surf.create_entity({name=ore, amount=math.random(dynamite_ore_min_amount,dynamite_ore_max_amount), position=pos})
						end
					end
				end
				
				table.remove( global.dynamites, i )
				i=i-1
			else
				dyna.countdown = dyna.countdown - 1
			end
			i=i+1
		end
	end
end

script.on_event(defines.events.on_tick, on_tick)

--------------------------------------------------------------------------------------
function is_fluid( name )
	local fluid
	for _, fluid in pairs(global.fluids) do
		if fluid == name then return true end
	end
	return false
end

--------------------------------------------------------------------------------------
function check_zone_used( inj, pos )
	local x_grid = math.floor( pos.x ) + 0.5
	local y_grid = math.floor( pos.y ) + 0.5
	
	-- test if zone already used
	
	local zoneok = true
	local zone, dx, dy, dist
	
	for _, zone in pairs( global.zones_used ) do
		dx = zone.x - x_grid
		dy = zone.y - y_grid
		dist = math.sqrt(dx*dx+dy*dy)
		if dist < injector_zone_radius * 1.7 then 
			zoneok = false
			break
		end
	end
	
	return zoneok
end

--------------------------------------------------------------------------------------
function update_ores_fluids()
	-- update global tables & recipes for ores and fluids
	debug_print( "update ores" )

	global.ores = {}
	global.fluids = {}

	for _,force in pairs(game.forces) do
		force.reset_technologies()
		force.reset_recipes()
	end

	-- mod base
	add_ore("coal", 0.7)
	add_ore("stone", 0.6 )
	add_ore("iron-ore", 1.0)
	add_ore("copper-ore", 1.1)
	add_fluid("crude-oil")
	
	-- mod dark matter
	add_ore("tenemut", 0.5)
	
	-- mod nucular
	add_ore("uranium-ore", 0.6)
	
	-- mod thorium
	add_ore("monazite-ore", 0.6)
	
	-- mod bobores / 5dim
	add_ore("bauxite-ore", 0.7)
	add_ore("cobalt-ore", 0.5)
	add_ore("gem-ore", 0.4)
	add_ore("gold-ore", 0.3)
	add_ore("lead-ore", 0.7)
	add_ore("nickel-ore", 0.7)
	add_ore("quartz", 0.5)
	add_ore("rutile-ore", 0.7)
	add_ore("silver-ore", 0.4)
	add_ore("tungsten-ore", 0.5)
	add_ore("tin-ore", 0.7)
	add_ore("zinc-ore", 0.7)
	
	-- mod crafted artifacts
	add_ore("rare-earth", 0.6)
	
	-- mod reactor
	add_ore("nuclear-ores", 0.6)
	add_fluid("Nature-Gas")
	
	debug_print( "ores=",#global.ores )
	debug_print( "fluids=",#global.fluids )
end

--------------------------------------------------------------------------------------
function add_ore(ore, proba)
	if game.entity_prototypes[ore] then  
		debug_print( "update ore ", ore )
		table.insert( global.ores, {ore, proba} )
		local force
		local dyna_name = "dynamite-" .. ore
		for _,force in pairs(game.forces) do
			if game.entity_prototypes[dyna_name] then  
				force.recipes[dyna_name].enabled = force.technologies["advanced-mining"].researched
				debug_print( "dyna recipe updated" )
			else
				debug_print( "dyna recipe not found" )
			end
		end
	end
end
	
--------------------------------------------------------------------------------------
function add_fluid(fluid, proba)
	if game.entity_prototypes[fluid] then  
		debug_print( "update fluid ", fluid )
		table.insert( global.fluids, fluid )
		local force
		local dyna_name = "dynamite-" .. fluid
		for _,force in pairs(game.forces) do
			force.recipes[dyna_name].enabled = force.technologies["advanced-mining"].researched
		end
	end
end
	
--------------------------------------------------------------------------------------
function scramble_ores()
	local ores2 = {}
	local n = #global.ores
	local m
	
	while n > 0 do
		m = math.random(1,#global.ores)
		table.insert( ores2, global.ores[m] )
		table.remove( global.ores, m )
		n=n-1
	end
	
	global.ores = ores2
end

--------------------------------------------------------------------------------------
function initiate_resources_zones( inj, pos )
	-- initialize ores zones (centers and zones)
	local x_grid = math.floor( pos.x ) + 0.5
	local y_grid = math.floor( pos.y ) + 0.5
	local angle = math.pi*2*math.random()
	local n, dist
	local add_zones = 0
	local larg = 0
	local ore
	
	for _, ore in pairs(global.ores) do
		larg = larg + ore[2]
	end
	
	larg = math.floor(injector_zone_radius * (9.5+math.random())/10 * (6+larg)/(6+3.1))
	
	scramble_ores()
	
	if debug_status then
		local s = "ores: "
		for _, ore in pairs(global.ores) do
			s = s .. ore[1] .. "(" .. ore[2] .. "),"
		end
		debug_print( s )
	end
		
	for n = 1, #global.ores+injector_add_zones do
		local n2
		if n <= #global.ores then
			n2 = n
		else
			add_zones = 1
			n2 = math.random(1,#global.ores)
		end
		
		dist = math.random(larg*(0.2+0.2*add_zones),larg*(0.63-0.6/#global.ores))
		angle = angle + math.pi * 2 / #global.ores * (4.5+math.random())/5
		
		table.insert(inj.centers ,
			{
				x = x_grid + dist * math.cos(angle),
				y = y_grid + dist * math.sin(angle),
				larg = min(larg*0.7, larg-dist) * (3+math.random())/4 * global.ores[n2][2] / (1+add_zones),
				invweight = 0.75+0.25*math.random()*(1-add_zones),
				ore = global.ores[n2][1],
			}
		)
	end
	
	debug_print( "centers=" .. #inj.centers )
	for n, center in pairs(inj.centers) do
		debug_print( "center" .. n .. "=" .. center.x .. "," .. center.y .. " " .. center.ore )
	end
	
	local x, y
	
	for y=y_grid-larg,y_grid+larg do
		for x=x_grid-larg,x_grid+larg do
			dist_min = 5 * larg
			n_min = -1
		
			for n, center in pairs(inj.centers) do
				dx = x - center.x
				dy = y - center.y
				dist = math.sqrt(dx*dx+dy*dy)
				
				if dist <= center.larg then
					if dist * center.invweight * (0.85+0.3*math.random()) < dist_min then
						dist_min = dist
						n_min = n
					end
				end
			end
			
			if n_min ~= -1 then
				center = inj.centers[n_min]
				
				table.insert(inj.zones, 
					{
						position = { x = x, y = y },
						ore = center.ore,
						amount_target = math.floor(math.random( injector_ore_min_amount, injector_ore_max_amount ) ),
						amount_divisor = 128,
						revealed = false,
						entity = nil,
					}
				)
			end
		end
	end
end

--------------------------------------------------------------------------------------
function reveal_resources_progressive( inj )
	local n_update = 0
	local zone
	local surf = inj.entity.surface
	
	-- increase already placed resources, and generate random explosions
	
	for _, zone in pairs(inj.zones) do
		if (not zone.revealed ) and zone.entity and zone.entity.valid then
			zone.entity.amount = max( 1, zone.amount_target / zone.amount_divisor )
			
			if zone.amount_divisor <= 1 then
				zone.revealed = true
			else
				zone.amount_divisor = zone.amount_divisor / 2
				n_update = n_update + 1
			end
			
			inj.n_boom = inj.n_boom - 1
			
			if inj.n_boom < 0 then
				surf.create_entity({name= "explosion", position=zone.position})
				inj.n_boom = math.random( 80 , 120 )
			end
		end
	end
	
	-- place all oils at the first call (not top interfere with further ores placements)
	local n_try = 100
	
	if inj.max_oils ~= 0 then
		debug_print( "max oils " .. inj.max_oils )
		while (inj.max_oils > 0) and (n_try > 0) do
			local n = math.random(1, #(inj.zones) )
			
			local pos = inj.zones[n].position
			local ents = surf.find_entities_filtered({area = square_area(pos,1.4), type = "resource"})

			if ents[1] == nil then				
				local am = 75 * math.random( injector_oil_min_yield, injector_oil_max_yield )
				local fluid = global.fluids[1+(inj.max_oils%(#global.fluids))]
				surf.create_entity({name=fluid, amount=am, position=pos})
				inj.max_oils = inj.max_oils - 1 
			end
				
			n_try = n_try - 1
		end
		
		inj.max_oils = 0
	end
	
	-- place some ores
	
	n_try = inj.nb_try_ores
	
	while n_try > 0  do
		-- choose a random zone
		n0 = math.random(1, #inj.zones )
		n = n0
		
		-- find next unrevealed zone
		while (inj.zones[n].entity ~= nil) or inj.zones[n].revealed do
			n = n + 1
			if n > #inj.zones then 
				n = 1 
			end
			
			if n == n0 then
				n = -1 
				break
			end
		end
		
		if n > 0 then
			zone = inj.zones[n]
			
			local pos = zone.position
			local ents = surf.find_entities_filtered({area = square_area(pos,0.4), type = "resource"})
			local tile = surf.get_tile(pos.x, pos.y)
			
			if ents[1] == nil and (tile.name ~= "water") and (tile.name ~= "deepwater") then
				zone.entity = surf.create_entity( {name=zone.ore, amount=1, position=pos} )
				n_update = n_update + 1
			else
				zone.revealed = true
			end			
		else
			inj.nb_try_ores = 0
			break
		end
		
		n_try = n_try - 1
	end

	if n_update == 0 then
		-- no more zones updated : end of process
		inj.status = status_finished
	end
end

--------------------------------------------------------------------------------------
function reveal_resources_once( inj )
	local n, zone
	local surf = inj.entity.surface
	
	for n, zone in pairs(inj.zones) do
		zone.entity = surf.create_entity({name=zone.ore, amount=zone.amount_target, position=zone.position})
	end
end

--------------------------------------------------------------------------------------

local interface = {}

function interface.reset( )
	debug_print( "reset" )
	
	update_ores_fluids()
	
	local player
	
	for _, player in pairs(game.players) do
		if player.gui.top.mntl_frame then
			player.gui.top.mntl_frame.destroy()
		end
	end
end

function interface.zap( )
	debug_print( "zap" )
	
	local pos = game.player.position
	local surf = game.player.surface
	local forc = game.player.force
	pos.x = pos.x + 3
	local ent = surf.create_entity({name="water-injector", amount=1, position=pos, force = forc})
	
	local inj = {
		entity = ent, 
		status = status_finished, 
		max_oils = math.floor(#global.fluids * injector_max_oils * (2+math.random())/3 ),
		cycles = 1,
		nb_try_ores = 0,
		centers={},
		zones={}, 
		elements={},
	}
	
	initiate_resources_zones( inj, ent.position )
	reveal_resources_once( inj )
	ent.active = false
end

remote.add_interface( "miningtools", interface )

-- /c remote.call( "miningtools", "zap" )
-- /c remote.call( "miningtools", "reset" )